class Errors:




    class IllegalType(Exception):

        def __init__(self,data_type,key,object={}):
            self.data_type = data_type
            self.key = key
            self.object = object
            super().__init__(f"Expected {data_type} for key {key} in object {object}")

    class MissingKey(Exception):

        def __init__(self,key,object={}):
            self.key = key
            self.object = object

        def __str__(self):
            return f"Missing key {self.key} in object {self.object}"


from .Errors import Errors
import uuid

class Schema:
    class InvalidSchema(Exception):
        def __init__(self, schema, key, value, type, required=False):
            self.schema = schema
            self.key = key
            self.value = value
            self.type = type
            self.required = required
            super().__init__(f"Invalid schema {schema}")

        def __str__(self):
            if self.required:
                return f"Invalid schema {self.schema} for key {self.key} with value {self.value} of type {type(self.type)} is required"
            return f"Invalid schema {self.schema} for key {self.key} with value {self.value} of type {type(self.value)} with expected type {self.type}"

    class InvalidIdFunction (Exception):
        def __init__(self, schema):
            self.schema = schema
            super().__init__(f"Invalid id function for schema {schema}")

        def __str__(self):
            return f"Invalid id function for schema {self.schema}"

    test_schema = {
        "key": {
            "type": str,
            "required": True
        }
    }

    def __init__(self, schema):
        self.schema = schema

    def has_id(self):
        return "_id" in self.schema

    def generate_id(self):
        return str(uuid.uuid4())

    def add_id(self):

        if not self.has_id():
            self.schema["_id"] = {
                "type": str,
                "required": True,
                "function": self.generate_id
            }
        if not "function" in self.schema["_id"]:
            self.schema["_id"]["function"] = self.generate_id

        if not callable(self.schema["_id"]["function"]):
            raise Schema.InvalidIdFunction(self.schema)





    def validate_schema(self, data):
        for key, value in self.schema.items():
            if key not in data:
                if not "required" in value or value["required"]:
                    raise Schema.InvalidSchema(self.schema, key, None, value["type"], True)
            else:
                if not isinstance(data[key], value["type"]):
                    raise Schema.InvalidSchema(self.schema, key, data[key], value["type"])
        return True

    def validate(self, key, value):
        if key not in self.schema:
            raise Errors.MissingKey(key, self.schema)
        if not isinstance(value, self.schema[key]["type"]):
            raise Errors.IllegalType(self.schema[key]["type"], key, value)
        return True


    def get_type(self, key):
        if key not in self.schema:
            raise Errors.MissingKey(key, self.schema)
        return self.schema[key]["type"]

    def is_required(self, key):
        if key not in self.schema:
            raise Errors.MissingKey(key, self.schema)
        return (not "required" in self.schema[key]) or (self.schema[key]["required"])

    def get_keys(self):
        return [key for key in self.schema]

    @staticmethod
    def gen_schema(data):
        schema = {}
        if data:
            for key, value in data.items():
                schema[key] = {
                    "type": type(value)
                }
        return schema


from .Schema import Schema


class DbObject:

    class IllegalDBTuple(Exception):

        req_size = 2

        def __init__(self, data):
            self.data = data
            super().__init__(f"Data must be a tuple, got {type(data)}")

        def __str__(self):
            if len(self.data) != self.req_size:
                return f"Data must be a tuple of length 2, got {len(self.data)}"
            return f"Data must be a tuple, got {type(self.data)}"

    class AttrNotFound(Exception):

        def __init__(self, key):
            self.key = key
            super().__init__(f"Attribute {key} not found")

        def __str__(self):
            return f"Attribute {self.key} not found"

    class IncompleteData(Exception):

        def __init__(self, key):
            self.key = key
            super().__init__(f"Key {key} is required")

        def __str__(self):
            return f"Key {self.key} is required"






    def __init__(self, schema, add_id=True):
        self.schema = Schema(schema) if isinstance(schema, dict) else schema
        if add_id:
            self.schema.add_id()
            self._id = self.schema.schema["_id"]["function"]()

    def __str__(self):
        return str(self.compile())

    def add(self, data):
        if not isinstance(data, tuple):
            raise DbObject.IllegalDBTuple(data)

        key = data[0]
        value = data[1]

        if self.schema.validate(key, value):
            setattr(self, key, value)
            return self

    def does_exist(self, key):
        return hasattr(self, key)

    def update(self,data):
        if not isinstance(data, tuple):
            raise DbObject.IllegalDBTuple(data)

        key = data[0]
        value = data[1]

        if self.schema.validate(key, value):
            if not self.does_exist(key):
                raise DbObject.AttrNotFound(key)
            self.add(data)
            return self

    def compile(self):
        data = {}
        for key, value in self.schema.schema.items():
            if not self.does_exist(key):
                if self.schema.is_required(key):
                    raise DbObject.IncompleteData(key)
            else:
                data[key] = getattr(self, key)
        return data

    def del_key(self, key):
        if self.does_exist(key): delattr(self, key)

    def req_keys_more(self):
        return self.keys_more(req=True)

    def keys_more(self,req=False):
        inc = []
        for key in self.schema.get_keys():
            #print(f"Debug: does_exist({key}) = {self.does_exist(key)}")
            if  self.schema.is_required(key) and (req and not self.does_exist(key)):
                inc.append(key)
        return inc
